@@ -12,9 +12,10 @@ package Net::SIP::StatelessProxy;
 use fields qw( dispatcher rewrite_contact nathelper force_rewrite );
 
 use Net::SIP::Util ':all';
-use Digest::MD5 qw(md5_hex md5);
+use Digest::MD5 qw(md5);
 use Carp 'croak';
 use List::Util 'first';
+use Hash::Util 'lock_ref_keys';
 use Net::SIP::Debug;
 
 ###########################################################################
@@ -28,6 +29,10 @@ use Net::SIP::Debug;
 #        if called on a string without @ which cannot rewritten back it
 #        should return undef. If not given a reasonable default will be
 #        used.
+#     rewrite_crypt: function(data,dir,add2mac) which will encrypt(dir>0) or
+#        decrypt(dir<0) data. Optional add2mac is added in MAC. Will return
+#        encrypted/decrypted data or undef if decryption failed because
+#        MAC did not match
 #     nathelper: Net::SIP::NAT::Helper used for rewrite SDP bodies.. (optional)
 #     force_rewrite: if true rewrite contact even if incoming and outgoing
 #         legs are the same
@@ -53,47 +58,55 @@ sub new {
 # default handler for rewriting, does simple XOR only,
 # this is not enough if you need to hide internal addresses
 sub _default_rewrite_contact {
-    my ($crypt,$disp,$contact,$leg_in,$leg_out) = @_;
+    my ($crypt,$disp,$contact,$leg_in,$leg_out,$force_rewrite) = @_;
 
-    if ( $contact =~m{\@} ) {
+    my $legdict;
+    my ($ileg_in,$ileg_out) = $disp->legs2i($leg_in,$leg_out,\$legdict);
+
+    if ($force_rewrite or $contact =~m{\@}) {
 	# needs to be rewritten - incorporate leg_in:leg_out
-	$contact = join("\|",
-	    (map { $_->key } ($leg_in,$leg_out)),
-	    $contact
-	);
-	# add 'r' in front of hex so it does not look like phone number
-	my $new = 'r'.unpack( 'H*',$crypt->($contact,1));
+	$contact = pack("nna*",$ileg_in,$ileg_out,$contact);
+	# add 'b' in front so it does not look like phone number
+	my $new = 'b'._encode_base32($crypt->($contact,1,$legdict));
 	DEBUG( 100,"rewrite $contact -> $new" );
 	return $new;
     }
 
-    if ( $contact =~m{^r([0-9a-f]+)$} ) {
+    if ( $contact =~m{^b([A-Z2-7]+)$} ) {
 	# needs to be written back
-	my $old = $crypt->(pack("H*",$1),-1) or do {
+	my $old = $crypt->(_decode_base32($1),-1,$legdict) or do {
 	    DEBUG(10,"no rewriting of $contact - bad encryption");
 	    return;
 	};
 	DEBUG(100,"rewrote back $contact -> $old");
-	(my $old_in,my $old_out,$old) = split( m{\|},$old,3);
-	my $new_in = $leg_in->key;
-	if ( $new_in ne $old_out ) {
-	    DEBUG(10,"no rewriting of $contact - went out through $old_out, came in through $new_in");
-	    return;
+	(my $iold_in,my $iold_out,$old) = unpack("nna*",$old);
+	if ($ileg_in ne $iold_out) {
+	    my ($old_out) = $disp->i2legs($iold_out);
+	    if ($leg_in->{contact} ne $old_out->{contact}
+		&& ! sip_uri_eq($leg_in->{contact},$old_out->{contact})) {
+		DEBUG(10,
+		    "no rewriting of %s - went out through %s, came in through %s",
+		    $contact, $old_out->{contact}, $leg_in->{contact});
+		return;
+	    }
 	}
 	if ( ref($leg_out) eq 'SCALAR' ) {
 	    # return the old_in as the new outgoing leg
-	    my @l = grep { $_->key eq $old_in } $disp->get_legs;
-	    if ( @l != 1 ) {
-		DEBUG(10,"no rewriting of $contact - cannot find leg $old_in");
+	    ($$leg_out) = $disp->i2legs($iold_in) or do {
+		DEBUG(10,"no rewriting of $contact - cannot find leg $iold_in");
 		return;
 	    }
-	    $$leg_out = $l[0];
-	} elsif ( $leg_out ) {
+	} elsif ($leg_out) {
 	    # check that it is the expected leg
-	    my $new_out = $leg_out->key;
-	    if ( $new_out ne $old_in ) {
-		DEBUG(10,"no rewriting of $contact - went in through $old_in, should got out through $new_out");
-		return;
+	    if ($ileg_out ne $iold_in) {
+		my ($old_in) = $disp->i2legs($iold_in);
+		if ($leg_out->{contact} ne $old_in->{contact}
+		    && ! sip_uri_eq($leg_out->{contact},$old_in->{contact})) {
+		    DEBUG(10,
+			"no rewriting of %s - went in through %s, should got out through %s",
+			$contact, $old_in->{contact}, $leg_out->{contact});
+		    return;
+		}
 	    }
 	}
 	DEBUG( 100,"rewrite back $contact -> $old" );
@@ -106,19 +119,35 @@ sub _default_rewrite_contact {
 }
 
 {
+    # This is only a simple implementation which is in no way cryptographic safe
+    # because it does use a broken cipher (RC4), pseudo-random keys and IV only
+    # and short keys. Nonetheless, it is probably safe for this purpose and does
+    # not depend on non-standard libs, but using openssl bindings might be both
+    # more secure and faster for this.
+    #
     # RC4 with seed + checksum, picks random key on first use
     # dir: encrypt(1),decrypt(-1), otherwise symmetric w/o seed and checksum
-    my @k;
+    my (@k,$mackey);
     sub _stupid_crypt {
-	my ($in,$dir) = @_;
-	@k = map { rand(256) } (0..20) if ! @k; # create random key 
+	my ($in,$dir,$add2mac) = @_;
+	$add2mac = '' if ! defined $add2mac;
+
+	if (!@k) {
+	    # create random key
+	    @k = map { rand(256) } (0..20);
+	    $mackey = pack("N",rand(2**32));
+	}
 
 	if ($dir>0) {
 	    $in = pack("N",rand(2**32)).$in;  # add seed
-	    $in .= substr(md5($in),0,4);      # add checksum
+	} else {
+	    # remove checksum and verify it
+	    my $cksum = substr($in,-4,4,'');
+	    substr(md5($in.$add2mac.$mackey),0,4) eq $cksum
+		or return;  # does not match
 	}
 
-	# RC4
+	# apply RC4 for encryption/decryption
 	my $out = '';
 	my @s = (0..255);
 	my $x = my $y = 0;
@@ -134,19 +163,41 @@ sub _default_rewrite_contact {
 	    $out .= pack('C',$_^=$s[($s[$x]+$s[$y])%256]);
 	}
 
-	if ( $dir<0 ) {
-	    my $cksum = substr($out,-4,4,'');           # remove checksum
-	    substr(md5($out),0,4) eq $cksum or return;  # verify it
-	    substr($out,0,4,'');                        # remove seed
+	if ($dir>0) {
+	    # add checksum
+	    $out .= substr(md5($out.$add2mac.$mackey),0,4);
+	} else {
+	    substr($out,0,4,'');  # remove seed
 	}
 	return $out;
     }
+
+    sub _encode_base32 {
+	my $data = shift;
+	$data = unpack('B*',$data);
+	my $text;
+	my $padsize =
+	$data .= '0' x ((5 - length($data) % 5) % 5); # padding
+	$data =~s{(.....)}{000$1}g;
+	$data = pack('B*',$data);
+	$data =~tr{\000-\037}{A-Z2-7};
+	return $data;
+    }
+
+    sub _decode_base32 {
+	my $data = shift;
+	$data =~ tr{A-Z2-7a-z}{\000-\037\000-\031};
+	$data = unpack('B*',$data);
+	$data =~s{...(.....)}{$1}g;
+	$data = substr($data,0,8*int(length($data)/8));
+	return pack('B*',$data);
+    }
 }
 
 ###########################################################################
-# handle incoming requests
+# handle incoming packets
 # Args: ($self,$packet,$leg,$from)
-#    $packet: Net::SIP::Request
+#    $packet: Net::SIP::Packet
 #    $leg: incoming leg
 #    $from: ip:port where packet came from
 # Returns: TRUE if packet was fully handled
@@ -248,22 +299,42 @@ sub __forward_response {
 	return;
     };
     my ($first,$param) = sip_hdrval2parts( via => $via );
-    my ($addr,$port) = $first =~m{([\w\-\.]+)(?::(\d+))?\s*$};
-    $port ||= 5060; # FIXME default for sip, not sips!
-    $addr = $param->{maddr} if $param->{maddr};
-    $addr = $param->{received} if $param->{received}; # where it came from
-    $port = $param->{rport} if $param->{rport}; # where it came from
-    @{ $entry->{dst_addr}} = ( "$addr:$port" );
-    DEBUG( 50,"get dst_addr from via header: $first -> $addr:$port" );
-
-    if ( $addr !~m{^[0-9\.]+$} ) {
-	$self->{dispatcher}->dns_host2ip(
-	    $addr,
-	    [ \&__forward_response_1,$self,$entry ]
-	);
-    } else {
-	__forward_response_1($self,$entry);
+    $first =~m{^SIP/\d\.\d(?:/(\S+))?\s+(.*)};
+    my $proto = lc($1) || 'udp';
+    my ($host,$port,$family) = ip_string2parts($2);
+    my $addr = $family && $host;
+    $port ||= $proto eq 'tls' ? 5061 : 5060;
+    if (my $alt_addr = $param->{received} || $param->{maddr}) {
+	my $alt_fam = ip_is_v46($alt_addr);
+	if ($alt_fam) {
+	    $addr = $alt_addr;
+	    $family = $alt_fam;
+	} else {
+	    DEBUG(10,"ignoring maddr/received because of invalid IP $alt_addr");
+	}
     }
+    $port = $param->{rport} if $param->{rport}; # where it came from
+    my $nexthop = lock_ref_keys({
+	proto  => $proto,
+	host   => $host || $addr,
+	addr   => $addr,
+	port   => $port,
+	family => $family
+    });
+    if ($addr) {
+	@{$entry->{dst_addr}} = $nexthop;
+	$DEBUG && DEBUG(50, "get dst_addr from via header: %s -> %s",
+	    $first, ip_parts2string($nexthop));
+	return __forward_response_1($self,$entry);
+    }
+
+    return $self->{dispatcher}->resolve_uri(
+	sip_sockinfo2uri($nexthop),
+	$entry->{dst_addr},
+	$entry->{outgoing_leg},
+	[ \&__forward_response_1,$self,$entry ],
+	undef,
+    );
 }
 
 ###########################################################################
@@ -274,17 +345,12 @@ sub __forward_response {
 sub __forward_response_1 {
     my Net::SIP::StatelessProxy $self = shift;
     my $entry = shift;
-    if ( @_ ) {
-	my ($errno,$ip) = @_;
-	unless ( $ip ) {
-	    DEBUG( 10,"cannot resolve address $entry->{dst_addr}[0]" );
-	    return;
-	}
-	# replace host part in dst_addr with ip
-	$entry->{dst_addr}[0] =~s{^(udp:|tcp:)?([^:]+)}{$1$ip};
+    if (@_) {
+	$DEBUG && DEBUG( 10,"cannot resolve address %s: @_",
+	    ip_parts2string($entry->{dst_addr}[0]));
+	return;
     }
-
-    __forward_packet_final( $self,$entry );
+    $self->__forward_packet_final($entry);
 }
 
 
@@ -315,9 +381,11 @@ sub __forward_request_getleg {
 	}
     } else {
 	my ($data,$param) = sip_hdrval2parts( route => $route );
-	my ($addr,$port) = $data =~m{([\w\-\.]+)(?::(\d+))?\s*$};
-	$port ||= 5060; # FIXME sips
-	my @legs = $self->{dispatcher}->get_legs(addr => $addr, port => $port);
+	my ($proto, $addr, $port, $family) =
+	    sip_uri2sockinfo($data, $param->{maddr} ? 1:0);
+	$port ||= $proto eq 'tls' ? 5061 : 5060;
+	my @legs = $self->{dispatcher}->get_legs(
+	    addr => $addr, port => $port, family => $family);
 	if ( ! @legs and $param->{maddr} ) {
 	    @legs = $self->{dispatcher}->get_legs( 
 		addr => $param->{maddr}, 
@@ -334,14 +402,9 @@ sub __forward_request_getleg {
     }
     if ( @route ) {
 	# still routing infos. Use next route as nexthop
-	my $route = $route[0] =~m{<([^\s>]+)>} && $1 || $route[0];
-	my ($data,$param) = sip_hdrval2parts( route => $route );
-	my ($addr,$port) = $data =~m{([\w\-\.]+)(?::(\d+))?\s*$};
-	$port ||= 5060; # FIXME sips
-	$entry->{nexthop} = $param->{maddr} 
-	    ? "$param->{maddr}:$port"
-	    : "$addr:$port";
-	DEBUG( 50, "setting nexthop from route $route to $entry->{nexthop}" );
+	my ($data,$param) = sip_hdrval2parts( route => $route[0] );
+	$entry->{nexthop} = $data;
+	DEBUG(50, "setting nexthop from route $route[0] to $entry->{nexthop}");
     }
 
     return $self->__forward_request_getdaddr($entry)
@@ -360,15 +423,14 @@ sub __forward_request_getdaddr {
     return __forward_request_1( $self,$entry )
 	if @{ $entry->{dst_addr}};
 
-    my $proto = $entry->{incoming_leg}{proto} eq 'tcp' ? [ 'tcp','udp' ]:undef;
     $entry->{nexthop} ||= $entry->{packet}->uri,
-    DEBUG( 50,"need to resolve $entry->{nexthop} proto=".( $proto ||'') );
+    DEBUG(50,"need to resolve $entry->{nexthop}");
     return $self->{dispatcher}->resolve_uri(
 	$entry->{nexthop},
 	$entry->{dst_addr},
 	$entry->{outgoing_leg},
 	[ \&__forward_request_1,$self,$entry ],
-	$proto,
+	undef,
     );
 }
 
@@ -380,6 +442,11 @@ sub __forward_request_1 {
     my Net::SIP::StatelessProxy $self = shift;
     my $entry = shift;
 
+    if (@_) {
+	DEBUG(10,"failed to resolve URI %s: @_",$entry->{nexthop});
+	return;
+    }
+
     my $dst_addr = $entry->{dst_addr};
     if ( ! @$dst_addr ) {
 	DEBUG( 10,"cannot find dst for uri ".$entry->{packet}->uri );
@@ -387,8 +454,8 @@ sub __forward_request_1 {
     }
     my %hostnames;
     foreach (@$dst_addr) {
-	my ($addr) = m{^(?:udp:|tcp:)?([^:]+)};
-	$hostnames{$addr} = undef if $addr !~m{^[0-9\.]+$};
+	ref($_) or Carp::confess("expected reference: $_");
+	$hostnames{$_->{host}} = $_->{host} if ! $_->{addr};
     }
     if ( %hostnames ) {
 	$self->{dispatcher}->dns_host2ip(
@@ -396,7 +463,7 @@ sub __forward_request_1 {
 	    [ \&__forward_request_2,$self,$entry ]
 	);
     } else {
-	__forward_packet_final($self,$entry);
+	$self->__forward_packet_final($entry);
     }
 }
 
@@ -412,17 +479,17 @@ sub __forward_request_2 {
     while ( my ($host,$ip) = each %$host2ip ) {
 	unless ( $ip ) {
 	    DEBUG( 10,"cannot resolve address $host" );
-	    @$dst_addr = grep { !m{^(?:\w*:)?\Q$host\E(?::)?} } @$dst_addr;
+	    @$dst_addr = grep { $_->{host} ne $host } @$dst_addr;
 	    next;
 	} else {
 	    DEBUG( 50,"resolved $host -> $ip" );
-	    s{^(\w*:)?\Q$host\E(:)?}{$1$ip$2} for (@$dst_addr);
+	    $_->{addr} = $ip for grep { $_->{host} eq $host } @$dst_addr;
 	}
     }
 
     return unless @$dst_addr; # nothing could be resolved
 
-    __forward_packet_final( $self,$entry );
+    $self->__forward_packet_final($entry);
 }
 
 
@@ -446,7 +513,7 @@ sub __forward_packet_final {
 	@$legs = ();
 	my @addr;
 	foreach my $addr (@$dst_addr) {
-	    my $leg = first { $_->can_deliver_to( $addr ) } @all_legs;
+	    my $leg = first { $_->can_deliver_to(%$addr) } @all_legs;
 	    if ( ! $leg ) {
 		DEBUG( 50,"no leg for $addr" );
 		next;
@@ -499,8 +566,8 @@ sub __forward_packet_final {
 	    # otherwise rewrite it
 	    } else {
 		$addr = invoke_callback($rewrite_contact,$addr,$incoming_leg,
-		    $outgoing_leg);
-		$addr .= '@'.$outgoing_leg->{addr}.':'.$outgoing_leg->{port};
+		    $outgoing_leg,1);
+		$addr .= '@'.$outgoing_leg->laddr(2);
 		my $cnew = sip_parts2hdrval( 'contact', $pre.$addr.$post, $p );
 		DEBUG( 50,"rewrote '$c' to '$cnew'" );
 		$c = $cnew;
@@ -562,7 +629,8 @@ sub do_nat {
     }
 
 
-    my $body = eval { $packet->sdp_body };
+    my $body = eval { $packet->cseq =~m{\b(?:INVITE|ACK)\b} 
+	&& $packet->sdp_body };
     if ( $@ ) {
 	DEBUG( 10, "malformed SDP body" );
 	return [ 500,"malformed SDP body" ];
@@ -591,21 +659,17 @@ sub do_nat {
     # Leg, delimited by "\0"
     my ($idfrom,$idto);
 
-    if ( my $from = $packet->get_header( 'from' ) ) {
-	my ($data,$param) = sip_hdrval2parts( from => $from );
-	my $tag = $param->{tag} || '';
-	$idfrom = "$data\0$tag";
-    } else {
-	return [ 0,'no FROM header in packet' ]
+    for([from => \$idfrom], [to => \$idto]) {
+	my ($k,$idref) = @$_;
+	if (my $v = $packet->get_header($k) ) {
+	    my ($uri,$param) = sip_hdrval2parts(from => $v);
+	    my ($dom,$user,$proto) = sip_uri2parts($uri);
+	    $$idref = "$proto:$user\@$dom\0".($param->{tag} || '');
+	} else {
+	    return [ 0,'no '.uc($k).' header in packet' ]
+	}
     }
 
-    if ( my $to = $packet->get_header( 'to' ) ) {
-	my ($data,$param) = sip_hdrval2parts( from => $to );
-	my $tag = $param->{tag} || '';
-	$idto = "$data\0$tag";
-    } else {
-	return [ 0,'no TO header in packet' ]
-    }
 
     # side is either 0 (request) or 1 (response)
     # If a request comes in 'from' points to the incoming_leg while
@@ -613,8 +677,8 @@ sub do_nat {
     # way around
 
     my $side;
-    my $ileg = join( ':', @{ $incoming_leg }{qw(addr port)} );
-    my $oleg = join( ':', @{ $outgoing_leg }{qw(addr port)} );
+    my $ileg = $incoming_leg->laddr(1);
+    my $oleg = $outgoing_leg->laddr(1);
     if ( $request ) {
 	$idfrom .= "\0".$ileg;
 	$idto   .= "\0".$oleg;
@@ -647,7 +711,7 @@ sub do_nat {
 	DEBUG( 100,"need to NAT SDP body: ".$body->as_string );
 
 	my $new_media = $nathelper->allocate_sockets(
-	    $callid,$cseq,$idfrom,$idto,$side,$outgoing_leg->{addr},
+	    $callid,$cseq,$idfrom,$idto,$side,$outgoing_leg->laddr(0),
 	    scalar( $body->get_media) );
 	if ( ! $new_media ) {
 	    DEBUG( 10,"allocation of RTP session failed for $callid|$cseq $idfrom|$idto|$side" );
